/**
 * The contents of this file are subject to the license and copyright detailed in the LICENSE and NOTICE files at the root of the source tree and available online at
 * http://www.askosi.org/license.html
 */
package be.destin.skos.analyze;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import be.destin.skos.core.Concept;
import be.destin.skos.core.SchemeInterface;
import be.destin.skos.core.SkosManager;

/**
 * This object collects information from documents in a query result set and stores how many times a concept is used.
 * After browsing the documents (or a sample of them), it is used to create frequencies maps which can be browsed using be.destin.skos.view.FaceterView to present a nice drill-down faceting panel to the user.
 * 
 * @author Christophe Dupriez
 *
 */
public class Faceter {
	

	/** log4j category */
	private static Logger log = Logger.getLogger(Faceter.class);

	private static final float hashTableLoadFactor = 0.75f;
	protected static int cacheSize = 0;
	private static final int defaultCacheSize = 100;
	
	private static LinkedHashMap<String/* query */, Faceter> entries = null;
	
	private HashMap<String,HashMap<String,Integer>> aboutMap = new HashMap<String,HashMap<String,Integer>>();
	private HashMap<String,HashMap<String,Integer>> conceptMap = new HashMap<String,HashMap<String,Integer>>();
    private HashMap<String,TreeMap<Integer,String[]>> freqMap = new HashMap<String,TreeMap<Integer,String[]>>();
    private HashMap<String,Integer> nbConcepts = new HashMap<String,Integer>();
    private HashMap<String,Integer> maxDepth = new HashMap<String,Integer>();
    private int nbDoc = 0;
    private int nbSample = 0;
    private int nbDocInBase = 0;
	private boolean analyzed = false;
    private int sampling = 0;
    
    private String query;

	private int menuLimit = 10;
    
	/**
	 * @return the Map that for each index (role) provides the Map that for each frequency provides the codes of the concepts with that frequency.  
	 */
	public HashMap<String, TreeMap<Integer, String[]>> getFreqMap() {
		return freqMap;
	}

	/**
	 * @return the number of documents in the query result set (may be a lot more than in the analyzed sample)
	 */
	public int getNbDoc() {
		return nbDoc;
	}

	/**
	 * @param nbDoc: size of the result set
	 */
	public void setNbDoc(int nbDoc) {
		this.nbDoc = nbDoc;
	}

	/**
	 * @param aRole : desired index (can use multiple conceptScheme)
	 * @return the number of different concepts referred by a query result set for a given index (role)
	 */
	public int getNbConcept(String aRole) {
		return nbConcepts.get(aRole);
	}

	/**
	 * @param aRole : desired index (can use multiple conceptScheme)
	 * @return the maximal depth in the thesaurus hierarchies including the different concepts referred by a query result set for a given index (role)
	 */
	public int getMaxDepth(String aRole) {
		Integer depth = maxDepth.get(aRole);
		if (depth == null) return 0;
		return depth;
	}

	/**
	 * @return if the Frequency Maps have already be generated starting from the data received about each document in the query (or a sample of them)
	 */
	public boolean isAnalyzed() {
		return analyzed;
	}

    /**
     * @return size of the sample of documents within the query result set which has been analysed
     */
    public int getNbSample() {
		return nbSample;
	}

	/**
	 * @param nbSample
	 */
	public void setNbSample(int nbSample) {
		this.nbSample = nbSample;
	}

	/**
	 * @return the sampling rate
	 */
	public int getSampling() {
		return sampling;
	}

	/**
	 * @param sampling: sets the sampling rate for this set (1 = every document)
	 */
	public void setSampling(int sampling) {
		this.sampling = sampling;
	}

	/**
	 * @return the search equation for this set
	 */
	public String getQuery() {
		return query;
	}

	/**
	 * @param query: the search equation for this result set
	 */
	public void setQuery(String query) {
		this.query = query;
	}
	/**
	 * @return
	 */
	public int getMenuLimit() {
		return menuLimit;
	}

	/**
	 * @param menuLimit
	 */
	public void setMenuLimit(int menuLimit) {
		this.menuLimit = menuLimit;
	}
    /**
     * Checks if the same query has not been analyzed recently and provides either the corresponding previous Faceter or a new empty one.
     * 
     * @param skosManager: ASKOSI configuration context
     * @param querystring: search equation
     * @return a previous Faceter or a new one (empty)
     */
    public static synchronized Faceter getInstance(SkosManager skosManager, String querystring) {
		if (entries == null || cacheSize == 0) {
			setCacheSize(defaultCacheSize);
		}
		String cacheKey = skosManager.getConfigDirectory()+SchemeInterface.fieldSeparator+querystring;
		Faceter faceter = entries.get(cacheKey);
		if (faceter == null) {
			faceter = new Faceter(querystring);
			entries.put(cacheKey, faceter);
		}
		return faceter;
	}

	/**
     * Always provides a new Faceter (and register it in the cache)
     * 
     * @param skosManager: ASKOSI configuration context
     * @param querystring: search equation
     * @return a new empty Faceter
	 */
	public static synchronized Faceter getFreshInstance(SkosManager skosManager, String querystring) {
		if (entries == null || cacheSize == 0) {
			setCacheSize(defaultCacheSize);
		}
		String cacheKey = skosManager.getConfigDirectory()+SchemeInterface.fieldSeparator+querystring;
		Faceter faceter = new Faceter(querystring);
		entries.put(cacheKey, faceter);
		return faceter;
	}

    /**
     * Checks if the same query has not been analyzed recently and provides the corresponding previous Faceter or "null" if none
     * 
     * @param skosManager: ASKOSI configuration context
     * @param querystring: search equation
     * @return a previous Faceter
     */
	public static synchronized Faceter getExistingInstance(SkosManager skosManager, String querystring) {
		if (entries == null || cacheSize == 0) return null;
		String cacheKey = skosManager.getConfigDirectory()+SchemeInterface.fieldSeparator+querystring;
		return entries.get(cacheKey);
	}
	
	/**
	 * Set the size for the Faceter objects cache. Can only grow.
	 * @param newCacheSize the new size of the Faceter cache
	 */
	public static synchronized void setCacheSize(int newCacheSize) {
		if (newCacheSize > Faceter.cacheSize) { // Cache can only grow!
			Faceter.cacheSize = newCacheSize;
			int hashTableCapacity = (int) Math.ceil(Faceter.cacheSize / hashTableLoadFactor) + 1;
			LinkedHashMap<String, Faceter> newEntries = new LinkedHashMap<String, Faceter>(hashTableCapacity,
					hashTableLoadFactor, true) {
				// An anonymous inner class to define a procedure which decides when cache is full
				private static final long serialVersionUID = 1;

				/*
				 * Manages the eldest entry when it has to be removed (cache full)
				 */
				protected boolean removeEldestEntry(Map.Entry<String,Faceter> eldest) {
					if (size() > Faceter.cacheSize) {
						//Faceter eldestFaceter = eldest.getValue();
						//if (eldestFaceter != null)
						//	removeMappings(eldestFaceter);
						return true;
					}
					else return false;
				}
			};
			if (entries != null) {
				newEntries.putAll(entries);
			}
			entries = newEntries;
		};
	}

	/**
	 * @return the cacheSize
	 */
	public static synchronized int getCacheSize() {
		return cacheSize;
	}


	protected Faceter (String queryKey) {
		super();
		this.query = queryKey;
	}

	/**
	 * Phase 1: Main method to call one document at a time, one index (role) at a time, one piece of content at a time.
	 * 
	 * @param skosManager: ASKOSI context
	 * @param docID: document ID (from 0, incremented by one when a new document is started)
	 * @param aRole: index (role, facet)
	 * @param content: string with the concept codes to count
	 */
	public void addField (SkosManager skosManager, int docID, String aRole, String content) {
//		log.info(content);
		if (docID >= nbSample) nbSample = docID+1;
		for (String aboutConcept : skosManager.toSetAbout(content)) {
			HashMap<String,Integer> thisMap = this.getAboutMap(aRole);
			Integer count = thisMap.get(aboutConcept);
			if (count == null) {
				thisMap.put(aboutConcept, Integer.valueOf(1));
			} else {
				thisMap.put(aboutConcept, Integer.valueOf(count.intValue()+1));
			}
		}
		/*
		for (Concept aConcept : skosManager.toSetScheme_Concept(content,null,false)) {
			HashMap<Concept,Integer> thisMap = this.getConceptMap(aRole);
			Integer count = thisMap.get(aConcept);
			if (count == null) {
				thisMap.put(aConcept, Integer.valueOf(1));
			} else {
				thisMap.put(aConcept, Integer.valueOf(count.intValue()+1));
			}
		}
		*/
	}

	/**
	 * @return the list of roles analyzed in this Faceter
	 */
	public Set<String> getFacets() {
		if (aboutMap == null) return null;
		return aboutMap.keySet();
	}

	/**
	 * Phase 2: Triggers the analysis (documents (sample) has been browsed)
	 * @param skosManager
	 */
	public synchronized void analyze(SkosManager skosManager) {
		if (analyzed) return;
		Set<String> roleEntries = aboutMap.keySet();
		for (String aRoleEntry : roleEntries) {
			HashMap<String,Integer> thisMap = this.getConceptMap(skosManager,aRoleEntry);
			HashMap<String,Integer> normMap = new HashMap<String,Integer>();
			for (Map.Entry<String, Integer> anEntry : thisMap.entrySet()) {
				//LinkedList<String> allBroaders = anEntry.getKey().getBroaderMatchTransitive();
				String aboutConcept = anEntry.getKey();
				Concept aConcept = skosManager.getScheme_Concept(aboutConcept);
				if (aConcept != null) {
					// Comme un concept est AUSSI index  son code principal, on NE compte PLUS les about qui ne sont pas principaux...
					String normAbout = aConcept.getScheme_About();
					if (normAbout.equals(aboutConcept)) {
						int depth = aConcept.getDepth();
						if (depth > 0) {
							int maxD = getMaxDepth(aRoleEntry);
							if (depth > maxD) {
								maxDepth.put(aRoleEntry,Integer.valueOf(depth));
							}
						}
						Integer freqObj = anEntry.getValue();
						if (freqObj != null && freqObj.intValue() > 0) {
							//log.debug(aboutConcept+"-->"+normAbout+"("+freqObj.intValue()+")");
							int freq = 0;
							Integer freqNorm = normMap.get(normAbout);
							if (freqNorm != null) {
								// Comme un concept est AUSSI index  son code principal, on NE compte PLUS les about qui ne sont pas principaux...
								// freq = freqObj.intValue() + freqNorm.intValue();
							}
							freq = freqObj.intValue();
							normMap.put(normAbout,freq);
							log.debug(aboutConcept+"-->"+normAbout+"("+freqObj.intValue()+"-->"+freq+")");
							anEntry.setValue(0);
							// thisMap.put(aboutConcept,0); NO REMOVE when iterating on an HashMap !!!
						}
					}
				} else {
					log.debug(aboutConcept+"-->NULL!");
				}
			}
			this.conceptMap.put(aRoleEntry,normMap);
			int nbConcept = 0;
			for (Entry<String, Integer> anEntry : normMap.entrySet()) {
				//LinkedList<String> allBroaders = anEntry.getKey().getBroaderMatchTransitive();
				String aboutConcept = anEntry.getKey();
				int freq = anEntry.getValue();
				if (freq > 0) {
					if (sampling > 1) {
						freq = freq * sampling;
						anEntry.setValue(freq);
					}
					putFreqMap(freq,aRoleEntry, aboutConcept);
					nbConcept++;
				}
			}
			nbConcepts.put (aRoleEntry,nbConcept);
		}
		analyzed = true;
	}

	/**
	 * @param aRole
	 * @param aboutConcept
	 * @param newFreq
	 */
	public synchronized void changeFreq(String aRole, String aboutConcept, int newFreq) {
		HashMap<String,Integer> thisMap = conceptMap.get(aRole);
		int curFreq = thisMap.get(aboutConcept);
		if (curFreq == newFreq) return;
		removeFreqMap(curFreq,aRole, aboutConcept);
		thisMap.put(aboutConcept, newFreq);
		putFreqMap(newFreq,aRole, aboutConcept);
	}

	/**
	 * Returns the Map of concepts for a role with their frequency (creates the map if it does not exist for that role)
	 * 
	 * @param aRole (index, facet) to map
	 * @return desired map
	 */
	public HashMap<String,Integer> getAboutMap(String aRole) {
		HashMap<String,Integer> thisMap = this.aboutMap.get(aRole);
		if (thisMap == null) {
			thisMap = new HashMap<String,Integer>();
			this.aboutMap.put(aRole, thisMap);
		}
		return thisMap;
	}


	/**
	 * @param skosManager
	 * @param aRole
	 * @return
	 */
	public HashMap<String,Integer> getConceptMap(SkosManager skosManager, String aRole) {
		HashMap<String,Integer> thisMap = this.conceptMap.get(aRole);
		if (thisMap == null) {
			thisMap = new HashMap<String,Integer>();
			this.conceptMap.put(aRole, thisMap);
			HashMap<String,Integer> thisAboutMap = this.aboutMap.get(aRole);
			if (thisAboutMap != null) {
				long t1 = System.currentTimeMillis();
				for (Entry<String,Integer> aboutEntry: thisAboutMap.entrySet()) {
					String aboutConcept = aboutEntry.getKey();
					if (aboutConcept != null && !aboutConcept.isEmpty()) {
						Integer prevCount = thisMap.get(aboutConcept);
						if (prevCount == null) prevCount = 0;
						thisMap.put(aboutConcept, prevCount+aboutEntry.getValue());
					}
				}
				log.info("Finding concepts for facet "+aRole+": "+(System.currentTimeMillis()-t1)+"ms for "+thisAboutMap.size()+" entries");
			}
		}
		return thisMap;
	}


	/**
	 * @param aRole
	 * @return
	 */
	public TreeMap<Integer,String[]> getFreqMap(String aRole) {
		TreeMap<Integer,String[]> thisMap = this.freqMap.get(aRole);
		if (thisMap == null) {
			thisMap = new TreeMap<Integer,String[]>();
			this.freqMap.put(aRole, thisMap);
		}
		return thisMap;
	}


	/**
	 * @param frequency
	 * @param aRole
	 * @param aboutConcept
	 */
	public void putFreqMap(int frequency, String aRole, String aboutConcept) {
		if (frequency == 0) return;
		if (aboutConcept == null || aboutConcept.isEmpty()) return;
		TreeMap<Integer,String[]> thisMap = this.getFreqMap(aRole); 
		String[] currConcepts = thisMap.get(frequency);
		String[] newConcepts;
		if (currConcepts == null) {
			newConcepts = new String[1];
			newConcepts[0] = aboutConcept;
		} else {
			newConcepts = currConcepts;
			boolean alreadyThere = false;
			for (int i=0; i<currConcepts.length; i++) {
				if (currConcepts[i].equals(aboutConcept)) {
					alreadyThere = true;
					break;
				}
			}
			if (!alreadyThere) {
				newConcepts = new String[currConcepts.length+1];
				for (int i=0; i<currConcepts.length; i++) {
					newConcepts[i] = currConcepts[i];
				}
				newConcepts[currConcepts.length] = aboutConcept;
			}
		}
		thisMap.put(frequency, newConcepts);
	}
	
	/**
	 * @param frequency
	 * @param aRole
	 * @param aboutConcept
	 */
	public void removeFreqMap(int frequency, String aRole, String aboutConcept) {
		if (frequency == 0) return;
		if (aboutConcept == null || aboutConcept.isEmpty()) return;
		TreeMap<Integer,String[]> thisMap = this.getFreqMap(aRole); 
		String[] currConcepts = thisMap.get(frequency);
		String[] newConcepts;
		if (currConcepts == null) {
			return;
		} else {
			newConcepts = currConcepts;
			boolean alreadyThere = false;
			for (int i=0; i<currConcepts.length; i++) {
				if (currConcepts[i].equals(aboutConcept)) {
					alreadyThere = true;
					break;
				}
			}
			if (alreadyThere) {
				if (currConcepts.length <= 1) {
					thisMap.remove(frequency);
				} else {
					newConcepts = new String[currConcepts.length-1];
					int j = 0;
					for (int i=0; i<currConcepts.length; i++) {
						if (!currConcepts[i].equals(aboutConcept)) {
							newConcepts[j] = currConcepts[i];
							j++;
						}
					}
					thisMap.put(frequency, newConcepts);
				}
			}
		}
	}

	public int getNbDocInBase() {
		return nbDocInBase;
	}

	public void setNbDocInBase(int nbDocInBase) {
		this.nbDocInBase = nbDocInBase;
	}

}
