/**
 * The contents of this file are subject to the license and copyright detailed in the LICENSE and NOTICE files at the root of the source tree and available online at
 * http://www.askosi.org/license.html
 */
package be.destin.skos.core;

import java.util.NavigableSet;
import java.util.TreeMap;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.bind.annotation.XmlType;

import org.apache.log4j.Logger;

import be.destin.skos.html.SkosHtmlManager;

@XmlAccessorType(XmlAccessType.PROPERTY)
@XmlType(propOrder={"searchURL","previewURL","completeURL","harvestURL","harvestFailure","harvestCount"})
public class SchemeUsage extends Count {
	
	/** log4j category */
	private static Logger log = Logger.getLogger(SchemeUsage.class);

	private NoScheme inScheme = null; // ConceptScheme or NotationScheme
	
	private String application = null;
	/*
	 * 3roundstones.com/vocab/vocab.pl: skos:prefLabel for role? label:plural and label:inverseSingular = qualified altLabel?
	 */
	private String searchURL;	// Brings the user to a search result based on current concept, application and role
	private String previewURL;	// Creates a preview (Internet Explorer Accelerators) on a search result based on current concept, application and role
    private String completeURL; // If the search result size is one and this URL is defined, this brings the user directly to display of an item
    private String harvestURL;	// Harvest count, main category, "purity" but also (eventually) counts in a time serie for a search based on current concept, application and role
    private String harvestFailure;	// Regular expression on the result to test for a failure
    private String harvestCount;	// Regular expression on the result to test for a success and extract the count OR XSLT to extract count, category, time serie?
    private boolean narrowerIncluded = false;
    private TreeMap<Integer,String[]> freqMap = null; // created by getFreqMap...

	public SchemeUsage() {
		super();
	}  
 
	public SchemeUsage(NoScheme aScheme, String anApplication, String aRole) {
		super();
		this.setInScheme(aScheme);
		this.application = anApplication;
		this.role = aRole;
	}  
 
	public SchemeUsage(NoScheme aScheme, String anApplication, String aRole, boolean narrowerIncluded) {
		super();
		this.setInScheme(aScheme);
		this.application = anApplication;
		this.role = aRole;
		this.narrowerIncluded = narrowerIncluded;
	}  
 
	public SchemeUsage(NoScheme aScheme, SchemeUsage model) {
		super(model);
		this.setInScheme(aScheme);
		// this.setCount(model.getCount());
		// this.role = model.role;
		this.application = model.getApplication();
		this.narrowerIncluded = model.narrowerIncluded;
		this.setSearchURL(model.getSearchURL());
		this.setPreviewURL(model.getPreviewURL());
		this.setCompleteURL(model.getCompleteURL());
		this.setHarvestURL(model.getHarvestURL());
		this.setHarvestFailure(model.getHarvestFailure());
		this.setHarvestCount(model.getHarvestCount());
	}  
 
	/**
	 * @return the completeURL
	 */
	public String getCompleteURL() {
		return completeURL;
	}
	/**
	 * @return the previewURL
	 */
	public String getPreviewURL() {
		return previewURL;
	}

	@XmlAttribute(required=true)
	public String getApplication() {
		return this.application;
	}

	public void setApplication(String key) {
		if (key == null) this.application=null;
		else this.application = key.trim();
	}

	/**
	 * @return the searchURL
	 */
	public String getSearchURL() {
		return searchURL;
	}


	/**
	 * @param completeURL the completeURL to set
	 */
	public void setCompleteURL(String completeURL) {
		this.completeURL = completeURL;
	}


	/**
	 * @param previewURL the previewURL to set
	 */
	public void setPreviewURL(String previewURL) {
		this.previewURL = previewURL;
	}


	/**
	 * @param searchURL the searchURL to set
	 */
	public void setSearchURL(String searchURL) {
		this.searchURL = searchURL;
	}

	/**
	 * @return the narrowerIncluded
	 */
	@XmlAttribute
	public boolean isPropagated() {
		return narrowerIncluded;
	}


	/**
	 * @param narrowerIncluded the narrowerIncluded to set
	 */
	public void setPropagated(boolean narrowerIncluded) {
		this.narrowerIncluded = narrowerIncluded;
	}


	/**
	 * @return the freqMap
	 */
	@XmlTransient
	public TreeMap<Integer,String[]> getFreqMap() {
		if (this.freqMap == null) {
			this.freqMap = new TreeMap<Integer,String[]>();
		}
		return this.freqMap;
	}


	/**
	 * @param freqMap the freqMap to set
	 */
	public void putFreqMap(int frequency, Concept aConcept) {
		if (aConcept == null) return;
		if (frequency == 0) return;
		String anAbout = aConcept.getAbout();
		String[] currConcepts = this.getFreqMap().get(frequency);
		String[] newConcepts;
		if (currConcepts == null) {
			newConcepts = new String[1];
			newConcepts[0] = anAbout;
		} else {
			newConcepts = currConcepts;
			boolean alreadyThere = false;
			for (int i=0; i<currConcepts.length; i++) {
				if (currConcepts[i].equals(anAbout)) {
					alreadyThere = true;
					break;
				}
			}
			if (!alreadyThere) {
				newConcepts = new String[currConcepts.length+1];
				for (int i=0; i<currConcepts.length; i++) {
					newConcepts[i] = currConcepts[i];
				}
				newConcepts[currConcepts.length] = anAbout;
			}
		}
		this.freqMap.put(frequency, newConcepts);
	}
	
	public String freqMapToHtml(SkosHtmlManager skosHtmlManager, int limit) {
		NavigableSet<Integer> freqs = this.getFreqMap().descendingKeySet();
		StringBuffer result = new StringBuffer();
		for (Integer aFreq : freqs) {
			result.append("<li>");
			result.append(aFreq.toString());
			result.append(':');
			if (limit <= 0) {
				result.append(" ...</li>");
				break;
			}
			boolean first = true;
			for (String anAbout: this.getFreqMap().get(aFreq)) {
				limit--;
				if (!first) {
					//result.append(',');
					result.append("<br/>");
				} else {
					first = false;
				}
				result.append(' ');
				Concept aConcept = this.inScheme.getConcept(anAbout);
				if (aConcept != null) {
					result.append(aConcept.toHtmlLabelUsage(skosHtmlManager));
				} else {
					result.append(this.inScheme.getAbout());
					result.append(SchemeInterface.separator);
					result.append(anAbout);
				}
			}
			result.append("</li>");
		}
		return result.toString();	
	}


	/**
	 * @return the inScheme
	 */
	@XmlTransient
	public NoScheme getInScheme() {
		return this.inScheme;
	}


	/**
	 * @param inScheme the inScheme to set
	 */
	public void setInScheme(NoScheme inScheme) {
		if (inScheme == null) {
			log.error("null Scheme for a SchemeUsage?");
		} else if (inScheme != null && inScheme instanceof ConceptScheme || inScheme instanceof NotationScheme) {
			this.inScheme = inScheme;
		} else {
			log.error(inScheme.getAbout()+" is not a ConceptScheme or a NotationScheme.");
		}
	}
	
	public String toString() {
		return "Usage of Scheme:"+inScheme.getAbout()+", application:"+application+", role:"+role+"="+count+(narrowerIncluded?"<":"");
	}

	public String getHarvestURL() {
		return harvestURL;
	}

	public void setHarvestURL(String harvestURL) {
		this.harvestURL = harvestURL;
	}

	public String getHarvestFailure() {
		return harvestFailure;
	}

	public void setHarvestFailure(String harvestFailure) {
		this.harvestFailure = harvestFailure;
	}

	public String getHarvestCount() {
		return harvestCount;
	}

	public void setHarvestCount(String harvestCount) {
		this.harvestCount = harvestCount;
	}
}
