/**
 * The contents of this file are subject to the license and copyright detailed in the LICENSE and NOTICE files at the root of the source tree and available online at
 * http://www.askosi.org/license.html
 */
package be.destin.skos.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.util.JAXBSource;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stax.StAXSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * This class provides a set of static methods to load and transform XML documents. It supports parameter-aware stylesheets (XSLT).
 * 
 * @author Christophe Dupriez
 *
 */
public class XMLUtil
{
    static Logger log = Logger.getLogger(XMLUtil.class);

	protected static HashMap<String, Templates> templateCache = new HashMap<String, Templates>();

    /**
     * Loads a W3C XML document from a file.
     * 
     * @param filename
     *            The name of the file to be loaded
     * @return a document object model object representing the XML file
     * @throws IOException
     * @throws ParserConfigurationException
     * @throws SAXException
     */
    public static Document loadXML(String filename) throws IOException,
            ParserConfigurationException, SAXException
    {
        DocumentBuilder builder = DocumentBuilderFactory.newInstance()
                .newDocumentBuilder();
        return builder.parse(new File(filename));
    }

    /**
     * @param inputFactory
     * @param xsltFilename
     * @param id
     * @return
     * @throws IOException
     * @throws TransformerConfigurationException
     * @throws XMLStreamException
     */
    public static Transformer openStreamingTransformer(XMLInputFactory inputFactory, String xsltFilename, String id) throws IOException,
            TransformerConfigurationException, XMLStreamException {
        
		Templates template = null;
		if (! Level.DEBUG.equals(log.getEffectiveLevel()) ) {
		    template = templateCache.get(xsltFilename); // Do not cache XSLT file when debugging!
		}
		if (template == null) {
			// rcupration de Templates
			File stylesheet = new File(xsltFilename);
			FileInputStream styleSheetStream = new FileInputStream(stylesheet);

    		XMLEventReader eventReaderXSL = null;
    		Source stylesource = null;

    		try
		    {
		        eventReaderXSL = inputFactory.createXMLEventReader(id,
		                new InputStreamReader(styleSheetStream));
		    }
		    catch (final javax.xml.stream.XMLStreamException e)
		    {
				log.error("Error opening XSL file", e);
				e.printStackTrace();
				throw e;
		    }
		    // define the Source object for the stylesheet
		    try
		    {
		    	stylesource = new StAXSource(eventReaderXSL);
		    }
		    catch (final javax.xml.stream.XMLStreamException e)
		    {
				log.error("Error opening XSL STAX source", e);
				e.printStackTrace();
				throw e;
		    }
			// cration de la fabrique
			TransformerFactory transfFactory = TransformerFactory.newInstance();
			template = transfFactory.newTemplates(stylesource);
			if (! Level.DEBUG.equals(log.getEffectiveLevel())) {
                templateCache.put (xsltFilename, template);
			}
		}
		return template.newTransformer();
    }

 /**
     * Applies a stylesheet to a given xml document.
     * 
     * @param xmlDocument
     *            the xml document to be transformed
     * @param xsltFilename
     *            the filename of the stylesheet
     * @return the transformed xml document
     * @throws Exception
     */
    public static Document transformDocument(Document xmlDocument,
            String xsltFilename) throws Exception
    {
        return transformDocument(xmlDocument, new Hashtable(), xsltFilename);
    }

    /**
     * Applies a stylesheet (that receives parameters) to a given xml document.
     * 
     * @param xmlDocument
     *            the xml document to be transformed
     * @param parameters
     *            the hashtable with the parameters to be passed to the
     *            stylesheet
     * @param xsltFilename
     *            the filename of the stylesheet
     * @return the transformed xml document
     * @throws Exception
     */
    public static Document transformDocument(Document xmlDocument,
            Hashtable parameters, String xsltFilename) throws Exception
    {

        // Generate a Transformer.
        Transformer transformer = openTransformer(xsltFilename);

        // set transformation parameters
        if (parameters != null)
        {
            Enumeration keys = parameters.keys();
            while (keys.hasMoreElements())
            {
                String key = (String) keys.nextElement();
                String value = (String) parameters.get(key);
                transformer.setParameter(key, value);
            }

        }

        // Create an empy DOMResult object for the output.
        DocumentBuilderFactory dFactory = DocumentBuilderFactory.newInstance();
        dFactory.setNamespaceAware(true);
        DocumentBuilder dBuilder = dFactory.newDocumentBuilder();
        Document dstDocument = dBuilder.newDocument();

        DOMResult domResult = new DOMResult(dstDocument);

        // Perform the transformation.
        transformer.transform(new DOMSource(xmlDocument), domResult);
        // Now you can get the output Node from the DOMResult.
        return dstDocument;
    }

    /**
     * Applies a stylesheet (that receives parameters) to a given xml document.
     * The resulting XML document is converted to a string after transformation.
     * 
     * @param xmlDocument
     *            the xml document to be transformed
     * @param parameters
     *            the hashtable with the parameters to be passed to the
     *            stylesheet
     * @param xsltFilename
     *            the filename of the stylesheet
     * @return the transformed xml document as a string
     * @throws Exception
     */
    public static String transformDocumentAsString(Document xmlDocument,
            Hashtable parameters, String xsltFilename) throws Exception
    {

        // Generate a Transformer.
        Transformer transformer = openTransformer(xsltFilename);

        // set transformation parameters
        if (parameters != null)
        {
            Enumeration keys = parameters.keys();
            while (keys.hasMoreElements())
            {
                String key = (String) keys.nextElement();
                String value = (String) parameters.get(key);
                transformer.setParameter(key, value);
            }

        }

        StringWriter stringWriter = new StringWriter();
        StreamResult streamResult = new StreamResult(stringWriter);

        // Perform the transformation.
        transformer.transform(new DOMSource(xmlDocument), streamResult);
        // Now you can get the output Node from the DOMResult.
        return stringWriter.toString();
    }

    /**
     * Applies a stylesheet (that receives parameters) to a given xml document.
     * The resulting XML document is written with provided writer after transformation.
     * 
     * @param xmlDocument
     *            the xml document to be transformed
     * @param parameters
     *            the hashtable with the parameters to be passed to the
     *            stylesheet
     * @param xsltFilename
     *            the filename of the stylesheet
     * @return the transformed xml document as a string
     * @throws Exception
     */
    public static void transformObject2Writer(String xsltFilename, 
            Hashtable<String,String> parameters, Object o, Writer out) throws Exception
    {
         if (o == null) log.error (xsltFilename+" applied on a null object for JAXB ?");
         else {
			// jaxbContext is a JAXBContext object from which 'o' is created.
			JAXBContext jaxbContext = JAXBContext.newInstance(o.getClass());
			JAXBSource source = new JAXBSource( jaxbContext, o );
                   
			Transformer transformer = openTransformer(xsltFilename);
           //         transformer.setOutputProperty(OutputKeys.INDENT, "yes");
           //         transformer.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1");	

			// set transformation parameters
			if (parameters != null)
			{
			    Enumeration<String> keys = parameters.keys();
			    while (keys.hasMoreElements())
			    {
			        String key = (String) keys.nextElement();
			        String value = (String) parameters.get(key);
			        transformer.setParameter(key, value);
			    }

			}

			StreamResult streamResult = new StreamResult(out);

			// Perform the transformation.
			transformer.transform(source, streamResult);
         }
	}

    /**
     * Applies a stylesheet to a given xml document.
     * 
     * @param xmlDocument
     *            the xml document to be transformed
     * @param xsltFilename
     *            the filename of the stylesheet
     * @return the transformed xml document
     * @throws Exception
     */
    public static String transformDocumentAsString(Document xmlDocument,
            String xsltFilename) throws Exception
    {
        // Generate a Transformer.
        Transformer transformer = openTransformer(xsltFilename);

        StringWriter stringWriter = new StringWriter();
        StreamResult streamResult = new StreamResult(stringWriter);

        // Perform the transformation.
        transformer.transform(new DOMSource(xmlDocument), streamResult);
        // Now you can get the output Node from the DOMResult.
        return stringWriter.toString();
    }

	/**
	 * @param xsltFilename
	 * @return
	 * @throws IOException
	 * @throws TransformerConfigurationException
	 * @throws SAXException
	 */
	public static Transformer openTransformer(String xsltFilename) throws IOException,
	        TransformerConfigurationException, SAXException {
	    
			Templates template = null;
			if (! Level.DEBUG.equals(log.getEffectiveLevel()) ) {
			    template = templateCache.get(xsltFilename); // Do not cache XSLT file when debugging!
			}
			if (template == null) {
				// rcupration de Templates
				File stylesheet = new File(xsltFilename);
				StreamSource stylesource = new StreamSource(stylesheet);
				// cration de la fabrique
				TransformerFactory fabrique = TransformerFactory.newInstance();
				template = fabrique.newTemplates(stylesource);
				if (! Level.DEBUG.equals(log.getEffectiveLevel())) {
	                templateCache.put (xsltFilename, template);
				}
			}
			return template.newTransformer();
	}

	/**
	 * Imposer le XSLT standard de JAVA:
	 * com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl
	 * et pas: -Djavax.xml.transform.TransformerFactory=net.sf.saxon.TransformerFactoryImpl
	 * et pas: -Djavax.xml.transform.TransformerFactory=org.apache.xalan.processor.TransformerFactoryImpl
	 * 
	 * @param xsltFilename
	 * @return
	 * @throws IOException
	 * @throws TransformerConfigurationException
	 * @throws SAXException
	 */
	public static Transformer openStandardTransformer(String xsltFilename) throws IOException,
	TransformerConfigurationException, SAXException {

		Templates template = null;
		if (! Level.DEBUG.equals(log.getEffectiveLevel()) ) {
			template = templateCache.get(xsltFilename); // Do not cache XSLT file when debugging!
		}
		if (template == null) {
			// rcupration de Templates
			File stylesheet = new File(xsltFilename);
			StreamSource stylesource = new StreamSource(stylesheet);
			// cration de la fabrique
			TransformerFactory fabrique = com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl.newInstance("com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl",null);
			template = fabrique.newTemplates(stylesource);
			if (! Level.DEBUG.equals(log.getEffectiveLevel())) {
				templateCache.put (xsltFilename, template);
			}
		}
		return template.newTransformer();
	}

	public final static byte[] begXmlDecl = new byte[]{ '<', '?', 'x', 'm', 'l' };

	public final static byte[] endXmlDecl = new byte[]{ '?', '>' };

}
